from django.core.management import call_command
from django.test import TestCase as DjangoTestCase

from django_cassandra_engine.utils import (
    get_cassandra_connections,
    get_cassandra_db_aliases,
)


class TestCase(DjangoTestCase):
    databases = list(get_cassandra_db_aliases())

    def _should_reload_connections(self):
        return False

    def _fixture_teardown(self):
        """
        Allow normal django TestCase fixture teardown, but also flush the test
        database for each cassandra alias.
        """
        super(TestCase, self)._fixture_teardown()

        for alias, _ in get_cassandra_connections():
            # Flush the database
            call_command(
                "flush",
                verbosity=1,
                interactive=False,
                database=alias,
                skip_checks=True,
                reset_sequences=False,
                allow_cascade=False,
                inhibit_post_migrate=True,
            )



import warnings

from django.core.validators import DecimalValidator
from django.utils.text import capfirst
from rest_framework import serializers
from rest_framework.serializers import (
    CharField,
    ChoiceField,
    ModelField,
    models,
    postgres_fields,
)
from rest_framework.utils.field_mapping import (
    NUMERIC_FIELD_TYPES,
    ClassLookupDict,
    needs_label,
    validators,
)

from ..compat import columns


class DjangoCassandraModelSerializer(serializers.ModelSerializer):
    serializer_field_mapping = {
        columns.Text: serializers.CharField,
        columns.UUID: serializers.CharField,
        columns.Integer: serializers.IntegerField,
        columns.TinyInt: serializers.IntegerField,
        columns.SmallInt: serializers.IntegerField,
        columns.BigInt: serializers.IntegerField,
        columns.VarInt: serializers.IntegerField,
        columns.Counter: serializers.IntegerField,
        columns.DateTime: serializers.DateTimeField,
        columns.Date: serializers.DateField,
        columns.Float: serializers.FloatField,
        columns.Double: serializers.FloatField,
        columns.Decimal: serializers.DecimalField,
        columns.Boolean: serializers.BooleanField,
        columns.DateTime: serializers.DateTimeField,
        columns.List: serializers.ListField,
    }

    def get_field_kwargs(self, field_name, model_field):
        """Creates a default instance of a basic non-relational field."""
        kwargs = {}
        validator_kwarg = list(model_field.validators)

        # The following will only be used by ModelField classes.
        # Gets removed for everything else.
        kwargs["model_field"] = model_field

        if model_field.verbose_name and needs_label(model_field, field_name):
            kwargs["label"] = capfirst(model_field.verbose_name)

        if model_field.help_text:
            kwargs["help_text"] = model_field.help_text

        max_digits = getattr(model_field, "max_digits", None)
        if max_digits is not None:
            kwargs["max_digits"] = max_digits

        decimal_places = getattr(model_field, "decimal_places", None)
        if decimal_places is not None:
            kwargs["decimal_places"] = decimal_places

        if isinstance(model_field, models.TextField):
            kwargs["style"] = {"base_template": "textarea.html"}

        if isinstance(model_field, models.AutoField) or not model_field.editable:
            # If this field is read-only, then return early.
            # Further keyword arguments are not valid.
            kwargs["read_only"] = True
            return kwargs

        if model_field.has_default or model_field.blank or model_field.null:
            kwargs["required"] = False

        if model_field.null and not isinstance(model_field, models.NullBooleanField):
            kwargs["allow_null"] = True

        if model_field.blank and (
            isinstance(model_field, models.CharField)
            or isinstance(model_field, models.TextField)
            or isinstance(model_field, columns.Text)
        ):
            kwargs["allow_blank"] = True

        if isinstance(model_field, models.FilePathField):
            kwargs["path"] = model_field.path

            if model_field.match is not None:
                kwargs["match"] = model_field.match

            if model_field.recursive is not False:
                kwargs["recursive"] = model_field.recursive

            if model_field.allow_files is not True:
                kwargs["allow_files"] = model_field.allow_files

            if model_field.allow_folders is not False:
                kwargs["allow_folders"] = model_field.allow_folders

        if model_field.choices:
            # If this model field contains choices, then return early.
            # Further keyword arguments are not valid.
            kwargs["choices"] = model_field.choices
            return kwargs

        # Our decimal validation is handled in the field code,
        # not validator code.
        # (In Django 1.9+ this differs from previous style)
        if isinstance(model_field, models.DecimalField) and DecimalValidator:
            validator_kwarg = [
                validator
                for validator in validator_kwarg
                if not isinstance(validator, DecimalValidator)
            ]

        # Ensure that max_length is passed explicitly as a keyword arg,
        # rather than as a validator.
        max_length = getattr(model_field, "max_length", None)
        if max_length is not None and (
            isinstance(model_field, models.CharField)
            or isinstance(model_field, models.TextField)
        ):
            kwargs["max_length"] = max_length
            validator_kwarg = [
                validator
                for validator in validator_kwarg
                if not isinstance(validator, validators.MaxLengthValidator)
            ]

        # Ensure that min_length is passed explicitly as a keyword arg,
        # rather than as a validator.
        min_length = next(
            (
                validator.limit_value
                for validator in validator_kwarg
                if isinstance(validator, validators.MinLengthValidator)
            ),
            None,
        )
        if min_length is not None and isinstance(model_field, models.CharField):
            kwargs["min_length"] = min_length
            validator_kwarg = [
                validator
                for validator in validator_kwarg
                if not isinstance(validator, validators.MinLengthValidator)
            ]

        # Ensure that max_value is passed explicitly as a keyword arg,
        # rather than as a validator.
        max_value = next(
            (
                validator.limit_value
                for validator in validator_kwarg
                if isinstance(validator, validators.MaxValueValidator)
            ),
            None,
        )
        if max_value is not None and isinstance(model_field, NUMERIC_FIELD_TYPES):
            kwargs["max_value"] = max_value
            validator_kwarg = [
                validator
                for validator in validator_kwarg
                if not isinstance(validator, validators.MaxValueValidator)
            ]

        # Ensure that max_value is passed explicitly as a keyword arg,
        # rather than as a validator.
        min_value = next(
            (
                validator.limit_value
                for validator in validator_kwarg
                if isinstance(validator, validators.MinValueValidator)
            ),
            None,
        )
        if min_value is not None and isinstance(model_field, NUMERIC_FIELD_TYPES):
            kwargs["min_value"] = min_value
            validator_kwarg = [
                validator
                for validator in validator_kwarg
                if not isinstance(validator, validators.MinValueValidator)
            ]

        # URLField does not need to include the URLValidator argument,
        # as it is explicitly added in.
        if isinstance(model_field, models.URLField):
            validator_kwarg = [
                validator
                for validator in validator_kwarg
                if not isinstance(validator, validators.URLValidator)
            ]

        # EmailField does not need to include the validate_email argument,
        # as it is explicitly added in.
        if isinstance(model_field, models.EmailField):
            validator_kwarg = [
                validator
                for validator in validator_kwarg
                if validator is not validators.validate_email
            ]

        # SlugField do not need to include the 'validate_slug' argument,
        if isinstance(model_field, models.SlugField):
            validator_kwarg = [
                validator
                for validator in validator_kwarg
                if validator is not validators.validate_slug
            ]

        # IPAddressField do not need to include the 'validate_ipv46_address'
        # argument,
        if isinstance(model_field, models.GenericIPAddressField):
            validator_kwarg = [
                validator
                for validator in validator_kwarg
                if validator is not validators.validate_ipv46_address
            ]

        if getattr(model_field, "unique", False):
            warnings.warn(
                "UniqueValidator is currently not supported "
                "in DjangoCassandraSerializer"
            )

        if validator_kwarg:
            kwargs["validators"] = validator_kwarg

        return kwargs

    def build_standard_field(self, field_name, model_field):
        """Create regular model fields."""
        field_mapping = ClassLookupDict(self.serializer_field_mapping)

        field_class = field_mapping[model_field]
        field_kwargs = self.get_field_kwargs(field_name, model_field)
        if "choices" in field_kwargs:
            # Fields with choices get coerced into `ChoiceField`
            # instead of using their regular typed field.
            field_class = self.serializer_choice_field
            # Some model fields may introduce kwargs that would not be valid
            # for the choice field. We need to strip these out.
            # Eg. models.DecimalField(max_digits=3, decimal_places=1,
            #                         choices=DECIMAL_CHOICES)

            valid_kwargs = {
                "read_only",
                "write_only",
                "required",
                "default",
                "initial",
                "source",
                "label",
                "help_text",
                "style",
                "error_messages",
                "validators",
                "allow_null",
                "allow_blank",
                "choices",
            }
            for key in list(field_kwargs.keys()):
                if key not in valid_kwargs:
                    field_kwargs.pop(key)

        if not issubclass(field_class, ModelField):
            # `model_field` is only valid for the fallback case of
            # `ModelField`, which is used when no other typed field
            # matched to the model field.
            field_kwargs.pop("model_field", None)

        if not issubclass(field_class, CharField) and not issubclass(
            field_class, ChoiceField
        ):
            # `allow_blank` is only valid for textual fields.
            field_kwargs.pop("allow_blank", None)

        if postgres_fields and isinstance(model_field, postgres_fields.ArrayField):
            # Populate the `child` argument on `ListField` instances generated
            # for the PostgrSQL specfic `ArrayField`.
            child_model_field = model_field.base_field
            child_field_class, child_field_kwargs = self.build_standard_field(
                "child", child_model_field
            )
            field_kwargs["child"] = child_field_class(**child_field_kwargs)

        return field_class, field_kwargs



import os

from django.apps import AppConfig as DjangoAppConfig

from .compat import columns


class CallableBool:
    """An boolean-like object that is also callable for backwards compatibility."""

    def __init__(self, value):
        self.value = value

    def __bool__(self):
        return self.value

    def __call__(self):
        return self.value

    def __nonzero__(self):  # Python 2 compatibility
        return self.value

    def __repr__(self):
        return "CallableBool(%r)" % self.value

    def __eq__(self, other):
        return self.value == other

    def __ne__(self, other):
        return self.value != other

    def __or__(self, other):
        return bool(self.value or other)

    def __hash__(self):
        return hash(self.value)


def has_default(self):
    return CallableBool(self.default is not None)


# monkey patch Column.has_default to be able to use function call too
columns.Column.has_default = property(has_default)

if os.getenv("CQLENG_ALLOW_SCHEMA_MANAGEMENT") is None:
    os.environ["CQLENG_ALLOW_SCHEMA_MANAGEMENT"] = "1"


class AppConfig(DjangoAppConfig):
    name = "django_cassandra_engine"

    def connect(self):
        from django_cassandra_engine.utils import get_cassandra_connections

        for _, conn in get_cassandra_connections():
            conn.connect()

    def import_models(self, *args, **kwargs):
        self.connect()
        return super().import_models(*args, **kwargs)



from django_cassandra_engine.utils import get_cassandra_connections

from .compat import (
    Cluster,
    CQLEngineException,
    PlainTextAuthProvider,
    Session,
    connection,
)


class Cursor(object):
    def __init__(self, connection):
        self.connection = connection

    def execute(self, *args, **kwargs):
        return self.connection.execute(*args, **kwargs)

    def close(self):
        pass

    def fetchmany(self, _):
        return []

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class FakeConnection(object):
    def commit(self):
        pass

    def rollback(self):
        pass

    def cursor(self):
        return Cursor(None)

    def close(self):
        pass


class CassandraConnection(object):
    def __init__(self, alias, **options):
        self.alias = alias
        self.hosts = options.get("HOST").split(",")
        self.keyspace = options.get("NAME")
        self.user = options.get("USER")
        self.password = options.get("PASSWORD")
        self.options = options.get("OPTIONS", {})
        self.cluster_options = self.options.get("connection", {})
        self.session_options = self.options.get("session", {})
        self.connection_options = {
            "lazy_connect": self.cluster_options.pop("lazy_connect", False),
            "retry_connect": self.cluster_options.pop("retry_connect", False),
            "consistency": self.cluster_options.pop("consistency", None),
        }
        if self.user and self.password and "auth_provider" not in self.cluster_options:
            self.cluster_options["auth_provider"] = PlainTextAuthProvider(
                username=self.user, password=self.password
            )

        self.default = (
            alias == "default"
            or len(list(get_cassandra_connections())) == 1
            or self.cluster_options.pop("default", False)
        )

        self.register()

    def register(self):
        try:
            connection.get_connection(name=self.alias)
        except CQLEngineException:
            if self.default:
                from cassandra.cqlengine import models

                models.DEFAULT_KEYSPACE = self.keyspace

            for option, value in self.session_options.items():
                setattr(Session, option, value)
            if "cloud" in self.cluster_options:
                cluster = Cluster(**self.cluster_options)
                session = cluster.connect()
                connection.register_connection(
                    self.alias, default=self.default, session=session
                )
            else:
                connection.register_connection(
                    self.alias,
                    hosts=self.hosts,
                    default=self.default,
                    cluster_options=self.cluster_options,
                    **self.connection_options
                )

    @property
    def cluster(self):
        return connection.get_cluster(connection=self.alias)

    @property
    def session(self):
        return connection.get_session(connection=self.alias)

    def commit(self):
        pass

    def rollback(self):
        pass

    def cursor(self):
        return Cursor(self)

    def execute(self, qs, *args, **kwargs):
        self.session.set_keyspace(self.keyspace)
        return self.session.execute(qs, *args, **kwargs)

    def close(self):
        """We would like to keep connection always open by default"""

    def unregister(self):
        """Unregister this connection"""
        connection.unregister_connection(self.alias)



from django.db.backends.base.features import BaseDatabaseFeatures


class CassandraDatabaseFeatures(BaseDatabaseFeatures):
    string_based_auto_field = True
    supports_long_model_names = False
    supports_microsecond_precision = True
    supports_transactions = False
    can_rollback_ddl = True
    uses_savepoints = False
    requires_rollback_on_dirty_transaction = False
    atomic_transactions = True



from django.core.management.commands.flush import Command as FlushCommand

from django_cassandra_engine.utils import get_engine_from_db_alias


class Command(FlushCommand):
    def handle_noargs(self, **options):
        engine = get_engine_from_db_alias(options["database"])
        if engine == "django_cassandra_engine":
            options.update({"interactive": False, "inhibit_post_migrate": True})

        return super(Command, self).handle_noargs(**options)

    def handle(self, **options):
        engine = get_engine_from_db_alias(options["database"])
        if engine == "django_cassandra_engine":
            options.update({"interactive": False, "inhibit_post_migrate": True})

        return super(Command, self).handle(**options)

    @staticmethod
    def emit_post_syncdb(verbosity, interactive, database):
        engine = get_engine_from_db_alias(database)
        if engine != "django_cassandra_engine":
            return FlushCommand.emit_post_syncdb(verbosity, interactive, database)

    @staticmethod
    def emit_post_migrate(verbosity, interactive, database):
        engine = get_engine_from_db_alias(database)
        if engine != "django_cassandra_engine":
            return FlushCommand.emit_post_migrate(verbosity, interactive, database)



import logging

from cassandra.cqlengine.connection import set_default_connection
from django.db.backends.base.creation import BaseDatabaseCreation

from django_cassandra_engine.utils import get_default_cassandra_connection

from ..compat import create_keyspace_simple, drop_keyspace

logger = logging.getLogger(__name__)


class CassandraDatabaseCreation(BaseDatabaseCreation):
    def create_test_db(self, verbosity=1, autoclobber=False, **kwargs):
        """
        Creates a test database, prompting the user for confirmation if the
        database already exists. Returns the name of the test database created.
        """
        # Don't import django.core.management if it isn't needed.
        from django.conf import settings
        from django.core.management import call_command

        self.connection.connect()
        default_alias = get_default_cassandra_connection()[0]

        # If using django-nose, its runner has already set the db name
        # to test_*, so restore it here so that all the models for the
        # live keyspace can be found.
        self.connection.connection.keyspace = self.connection.settings_dict["NAME"]
        test_database_name = self._get_test_db_name()
        # Set all models keyspace to the test keyspace
        self.set_models_keyspace(test_database_name)

        if verbosity >= 1:
            test_db_repr = ""
            if verbosity >= 2:
                test_db_repr = " ('%s')" % test_database_name
            logger.info(
                "Creating test database for alias '%s'%s..."
                % (self.connection.alias, test_db_repr)
            )

        options = self.connection.settings_dict.get("OPTIONS", {})

        # temporarily enable schema metadata for sync_cassandra
        connection_options_copy = options.get("connection", {}).copy()
        if not connection_options_copy.get("schema_metadata_enabled", True):
            options["connection"]["schema_metadata_enabled"] = True
            self.connection.reconnect()
            set_default_connection(default_alias)

        replication_opts = options.get("replication", {})
        replication_factor = replication_opts.pop("replication_factor", 1)

        create_keyspace_simple(
            test_database_name,
            replication_factor,
            connections=[self.connection.alias],
        )

        settings.DATABASES[self.connection.alias]["NAME"] = test_database_name
        self.connection.settings_dict["NAME"] = test_database_name

        self.connection.reconnect()
        set_default_connection(default_alias)

        # Report syncdb messages at one level lower than that requested.
        # This ensures we don't get flooded with messages during testing
        # (unless you really ask to be flooded)
        call_command(
            "sync_cassandra",
            verbosity=max(verbosity - 1, 0),
            database=self.connection.alias,
        )

        # restore the original connection options
        if not connection_options_copy.get("schema_metadata_enabled", True):
            logger.info(
                "Disabling metadata on %s" % self.connection.settings_dict["NAME"]
            )
            options["connection"]["schema_metadata_enabled"] = connection_options_copy[
                "schema_metadata_enabled"
            ]
            self.connection.reconnect()
            set_default_connection(default_alias)

        return test_database_name

    def _destroy_test_db(self, test_database_name, verbosity=1, **kwargs):

        drop_keyspace(test_database_name, connections=[self.connection.alias])

    def set_models_keyspace(self, keyspace):
        """Set keyspace for all connection models"""
        for models in self.connection.introspection.cql_models.values():
            for model in models:
                model.__keyspace__ = keyspace



import subprocess

from django.db.backends.base.client import BaseDatabaseClient


class CassandraDatabaseClient(BaseDatabaseClient):
    executable_name = "cqlsh"

    def runshell(self, *args, **kwargs):
        settings_dict = self.connection.settings_dict
        args = [self.executable_name]
        if settings_dict["HOST"]:
            args.extend([settings_dict["HOST"].split(",")[0]])
        if settings_dict["PORT"]:
            args.extend([str(settings_dict["PORT"])])
        if settings_dict["USER"]:
            args += ["-u", settings_dict["USER"]]
        args += ["-k", settings_dict["NAME"]]

        subprocess.call(args)



from django.apps import apps
from django.core.management.base import BaseCommand, CommandError
from django.db import connections

from django_cassandra_engine.models import DjangoCassandraModel
from django_cassandra_engine.utils import get_engine_from_db_alias

from ...compat import Model, management


class Command(BaseCommand):
    help = "Sync Cassandra database(s)"

    def add_arguments(self, parser):
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=None,
            help="Nominates a database to synchronize.",
        )

    @staticmethod
    def _import_management():
        """
        Import the 'management' module within each installed app, to register
        dispatcher events.
        """
        from importlib import import_module

        for app_config in apps.get_app_configs():
            try:
                import_module(".management", app_config.name)
            except SystemError:
                # We get SystemError if INSTALLED_APPS contains the
                # name of a class rather than a module
                pass
            except ImportError as exc:
                # This is slightly hackish. We want to ignore ImportErrors
                # if the "management" module itself is missing -- but we don't
                # want to ignore the exception if the management module exists
                # but raises an ImportError for some reason. The only way we
                # can do this is to check the text of the exception. Note that
                # we're a bit broad in how we check the text, because different
                # Python implementations may not use the same text.
                # CPython uses the text "No module named management"
                # PyPy uses "No module named myproject.myapp.management"
                msg = exc.args[0]
                if not msg.startswith("No module named") or "management" not in msg:
                    raise

    def sync(self, alias):
        engine = get_engine_from_db_alias(alias)

        if engine != "django_cassandra_engine":
            raise CommandError("Database {} is not cassandra!".format(alias))

        connection = connections[alias]
        connection.connect()
        options = connection.settings_dict.get("OPTIONS", {})
        keyspace = connection.settings_dict["NAME"]
        replication_opts = options.get("replication", {})
        strategy_class = replication_opts.pop("strategy_class", "SimpleStrategy")
        replication_factor = replication_opts.pop("replication_factor", 1)

        self.stdout.write(
            "Creating keyspace {} [CONNECTION {}] ..".format(keyspace, alias)
        )

        if strategy_class == "SimpleStrategy":
            management.create_keyspace_simple(
                keyspace, replication_factor, connections=[alias]
            )
        else:
            management.create_keyspace_network_topology(
                keyspace, replication_opts, connections=[alias]
            )

        connection.connection.cluster.refresh_schema_metadata()
        connection.connection.cluster.schema_metadata_enabled = True

        for app_name, app_models in connection.introspection.cql_models.items():
            for model in app_models:
                self.stdout.write("Syncing %s.%s" % (app_name, model.__name__))
                # patch this object used for type check in management.sync_table()
                management.Model = (Model, DjangoCassandraModel)
                management.sync_table(model, keyspaces=[keyspace], connections=[alias])

    def handle(self, **options):

        self._import_management()

        database = options.get("database")
        if database is not None:
            return self.sync(database)

        cassandra_alias = None
        for alias in connections:
            engine = get_engine_from_db_alias(alias)
            if engine == "django_cassandra_engine":
                self.sync(alias)
                cassandra_alias = alias

        if cassandra_alias is None:
            raise CommandError(
                "Please add django_cassandra_engine backend to DATABASES!"
            )



from django.db.backends.base.operations import BaseDatabaseOperations


class CassandraDatabaseOperations(BaseDatabaseOperations):
    def pk_default_value(self):
        """
        Returns None, to be interpreted by back-ends as a request to
        generate a new key for an "inserted" object.
        """
        return None

    def quote_name(self, name):
        """
        Does not do any quoting, as it is not needed for most NoSQL
        databases.
        """
        return name

    def prep_for_like_query(self, value):
        """Does no conversion, parent string-cast is SQL specific."""
        return value

    def prep_for_iexact_query(self, value):
        """Does no conversion, parent string-cast is SQL specific."""
        return value

    def sql_flush(self, style, tables, *args, **kwargs):
        """
        Truncate all existing tables in current keyspace.
        :returns: an empty list
        """
        if tables:
            cql_list = []

            for table in tables:
                cql_list.append(f"TRUNCATE {table}")
            return cql_list
        return []

    def execute_sql_flush(self, *args):
        # In previous django versions first parameter was `using` and second `sql_list`.
        # In Django 3.1 only one parameter `sql_list` is present.
        # One thing is certain though: last parameter is `sql_list`
        for cql in args[-1]:
            self.connection.connection.execute(cql)

    def prepare_sql_script(self, sql):
        return [sql]



from django.db.backends.base.schema import BaseDatabaseSchemaEditor


class CassandraDatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    def prepare_default(self, value):
        return value

    def quote_value(self, value):
        return value

    def create_model(self, model):
        # Hack to disable checking migrations for cassandra connection
        try:
            from django.db.migrations.exceptions import MigrationSchemaMissing as Exc
        except ImportError:
            from django.core.exceptions import ImproperlyConfigured as Exc
        raise Exc("No schema for cassandra database")

    def delete_model(self, model):
        pass



import threading

from django_cassandra_engine.base.client import CassandraDatabaseClient
from django_cassandra_engine.base.creation import CassandraDatabaseCreation
from django_cassandra_engine.base.features import CassandraDatabaseFeatures
from django_cassandra_engine.base.introspection import CassandraDatabaseIntrospection
from django_cassandra_engine.base.operations import CassandraDatabaseOperations
from django_cassandra_engine.base.schema import CassandraDatabaseSchemaEditor
from django_cassandra_engine.base.validation import CassandraDatabaseValidation
from django_cassandra_engine.connection import CassandraConnection, FakeConnection
from django_cassandra_engine.utils import CursorWrapper

try:
    from django.db.backends.base.base import BaseDatabaseWrapper, connection_created
except ImportError:
    try:
        from django.db.backends import BaseDatabaseWrapper, connection_created
    except ImportError:
        from django.db.backends.signals import connection_created


class Database(object):
    class Error(Exception):
        pass

    class InterfaceError(Error):
        pass

    class DatabaseError(Error):
        pass

    class DataError(DatabaseError):
        pass

    class OperationalError(DatabaseError):
        pass

    class IntegrityError(DatabaseError):
        pass

    class InternalError(DatabaseError):
        pass

    class ProgrammingError(DatabaseError):
        pass

    class NotSupportedError(DatabaseError):
        pass


class DatabaseWrapper(BaseDatabaseWrapper):

    Database = Database
    vendor = "cassandra"

    operators = {
        "exact": "= %s",
        "contains": "CONTAINS %s",
        "gt": "> %s",
        "gte": ">= %s",
        "lt": "< %s",
        "lte": "<= %s",
    }

    client = None
    # Classes instantiated in __init__().
    client_class = CassandraDatabaseClient
    creation_class = CassandraDatabaseCreation
    features_class = CassandraDatabaseFeatures
    introspection_class = CassandraDatabaseIntrospection
    ops_class = CassandraDatabaseOperations
    validation_class = CassandraDatabaseValidation

    def __init__(self, *args, **kwargs):
        super(DatabaseWrapper, self).__init__(*args, **kwargs)

        self.commit_on_exit = False
        self.connected = False
        self.autocommit = True
        self._connect_lock = threading.RLock()

        if self.client is None:
            # Set up the associated backend objects
            self.features = CassandraDatabaseFeatures(self)
            self.ops = CassandraDatabaseOperations(self)
            self.client = CassandraDatabaseClient(self)
            self.creation = CassandraDatabaseCreation(self)
            self.validation = CassandraDatabaseValidation(self)
            self.introspection = CassandraDatabaseIntrospection(self)

        del self.connection

    def get_connection_params(self):
        return {}

    def get_new_connection(self, conn_params):
        return FakeConnection()

    def init_connection_state(self):
        pass

    def _set_autocommit(self, autocommit):
        pass

    def connect(self):
        with self._connect_lock:
            if not self.connected or self.connection is None:
                settings = self.settings_dict
                self.connection = CassandraConnection(self.alias, **settings)
                connection_created.send(sender=self.__class__, connection=self)
                self.connected = True

    def __getattr__(self, attr):
        if attr == "connection":
            assert not self.connected
            self.connect()
            return getattr(self, attr)
        raise AttributeError(attr)

    def reconnect(self):
        with self._connect_lock:
            if self.connected:
                self.connection.unregister()
                del self.connection
                self.connected = False
            self.connect()

    def close_if_unusable_or_obsolete(self):
        self.connect()

    def close(self):
        pass

    def _commit(self):
        pass

    def _rollback(self):
        pass

    def _cursor(self, *args, **kwargs):
        keyspace = self.settings_dict["NAME"]
        if (
            not self.connection.cluster.schema_metadata_enabled
            and keyspace not in self.connection.cluster.metadata.keyspaces
        ):
            self.connection.cluster.refresh_schema_metadata()

        self.connection.cluster.metadata.keyspaces[keyspace]
        return CursorWrapper(self.connection.cursor(), self)

    def schema_editor(self, *args, **kwargs):
        """Returns a new instance of this backends' SchemaEditor (Django>=1.7)"""
        return CassandraDatabaseSchemaEditor(self, *args, **kwargs)



"""Handle package compatibility."""

try:
    from dse import cqlengine
    from dse.auth import PlainTextAuthProvider
    from dse.cluster import Cluster, Session
    from dse.cqlengine import CQLEngineException, columns, connection, management, query
    from dse.cqlengine.management import create_keyspace_simple, drop_keyspace
    from dse.cqlengine.models import (
        BaseModel,
        ColumnDescriptor,
        Model,
        ModelDefinitionException,
        ModelException,
        ModelMetaClass,
    )
    from dse.util import OrderedDict
except ImportError:
    try:
        from cassandra import cqlengine
        from cassandra.auth import PlainTextAuthProvider
        from cassandra.cluster import Cluster, Session
        from cassandra.cqlengine import (
            CQLEngineException,
            columns,
            connection,
            management,
            query,
        )
        from cassandra.cqlengine.management import create_keyspace_simple, drop_keyspace
        from cassandra.cqlengine.models import (
            BaseModel,
            ColumnDescriptor,
            Model,
            ModelDefinitionException,
            ModelException,
            ModelMetaClass,
        )
        from cassandra.util import OrderedDict
    except ImportError:
        raise ImportError(
            "You must install either dse-driver or " + "cassandra-driver!"
        )



ORDER_BY_WARN = (
    '.order_by() with column "{col}" failed! '
    "Falling back to ordering in python. \n"
    "Exception was: {exc}"
)
ORDER_BY_ERROR_HELP = (
    "To enable fallback order_by() python implementation,\n"
    'set "CASSANDRA_FALLBACK_ORDER_BY_PYTHON = True" in Django Settings.'
)
PK_META_MISSING_HELP = (
    "\nOn Django Cassandra Models with more than one primary_key field,\n"
    "The model {} must specify class Meta attribute 'get_pk_field'.\n"
    "E.g.\n"
    "class Meta:\n"
    "  get_pk_field='id'\n"
    "So that Django knows which primary key field to use in queryies "
    "such as Model.objects.get(pk=123)"
)



from django.contrib.sessions.backends import cached_db
from django.contrib.sessions.backends.cached_db import SessionStore as CachedStore

from django_cassandra_engine.sessions.backends.db import SessionStore as DBStore
from django_cassandra_engine.sessions.models import Session as CassandraSession

cached_db.Session = CassandraSession

KEY_PREFIX = "dce.sessions.cached_db"


class SessionStore(DBStore, CachedStore):
    """Implements cached, cassandra backed sessions."""

    cache_key_prefix = KEY_PREFIX



import inspect

from django.conf import settings
from django.db import DEFAULT_DB_ALIAS

from .compat import cqlengine


class CursorWrapper(object):
    """Simple CursorWrapper implementation based on `django.db.utils.CursorWrapper`"""

    def __init__(self, cursor, db):
        self.cursor = cursor
        self.db = db

    WRAP_ERROR_ATTRS = frozenset(["fetchone", "fetchmany", "fetchall", "nextset"])

    def __getattr__(self, attr):
        cursor_attr = getattr(self.cursor, attr)
        if attr in CursorWrapper.WRAP_ERROR_ATTRS:
            return self.db.wrap_database_errors(cursor_attr)
        else:
            return cursor_attr

    def __iter__(self):
        return iter(self.cursor)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def callproc(self, procname, params=None):
        with self.db.wrap_database_errors:
            if params is None:
                return self.cursor.callproc(procname)
            else:
                return self.cursor.callproc(procname, params)

    def execute(self, sql, params=None):
        with self.db.wrap_database_errors:
            if params is None:
                return self.cursor.execute(sql)
            else:
                return self.cursor.execute(sql, params)

    def executemany(self, sql, param_list):
        with self.db.wrap_database_errors:
            return self.cursor.executemany(sql, param_list)


def get_installed_apps():
    """Return list of all installed apps"""
    from django.apps import apps

    return [
        a.models_module for a in apps.get_app_configs() if a.models_module is not None
    ]


def get_cql_models(app, connection=None, keyspace=None):
    """
    :param app: django models module
    :param connection: connection name
    :param keyspace: keyspace
    :return: list of all cassandra.cqlengine.Model within app that should be
    synced to keyspace.
    """
    from .models import DjangoCassandraModel

    models = []
    single_cassandra_connection = len(list(get_cassandra_connections())) == 1
    is_default_connection = (
        connection == DEFAULT_DB_ALIAS or single_cassandra_connection
    )

    for name, obj in inspect.getmembers(app):
        cql_model_types = (cqlengine.models.Model, DjangoCassandraModel)
        if (
            inspect.isclass(obj)
            and issubclass(obj, cql_model_types)
            and not obj.__abstract__
        ):
            if (
                obj.__connection__ == connection
                or (obj.__connection__ is None and is_default_connection)
                or obj.__connection__ is None
                and obj.__keyspace__ is not None
                and obj.__keyspace__ == keyspace
            ):
                models.append(obj)

    return models


def get_cassandra_connections():
    """
    :return: List of tuples (db_alias, connection) for all cassandra
    connections in DATABASES dict.
    """
    from django.db import connections

    for alias in connections:
        engine = connections[alias].settings_dict.get("ENGINE", "")
        if engine == "django_cassandra_engine":
            yield alias, connections[alias]


def get_default_cassandra_connection():
    """
    Return first default cassandra connection
    :return:
    """
    for alias, conn in get_cassandra_connections():
        if conn.connection.default:
            return alias, conn

    return list(get_cassandra_connections())[0]


def get_cassandra_connection(alias=None, name=None):
    """:return: cassandra connection matching alias or name or just first found."""
    for _alias, connection in get_cassandra_connections():
        if alias is not None:
            if alias == _alias:
                return connection
        elif name is not None:
            if name == connection.settings_dict["NAME"]:
                return connection
        else:
            return connection


def get_cassandra_db_aliases():
    from django.db import connections

    for alias in connections:
        engine = connections[alias].settings_dict.get("ENGINE", "")
        if engine == "django_cassandra_engine":
            yield alias


def get_cassandra_db_alias():
    return get_cassandra_db_aliases().__next__()


def get_engine_from_db_alias(db_alias):
    """
    :param db_alias: database alias
    :return: database engine from DATABASES dict corresponding to db_alias
             or None if db_alias was not found
    """
    return settings.DATABASES.get(db_alias, {}).get("ENGINE", None)



from django_cassandra_engine.management.commands import sync_cassandra
from django_cassandra_engine.management.commands.sync_cassandra import (
    Command as SyncCommand,
)
from django_cassandra_engine.utils import get_engine_from_db_alias


class Command(SyncCommand):
    def handle_noargs(self, **options):
        engine = get_engine_from_db_alias(options["database"])

        # Call regular syncdb if engine is different from ours
        if engine != "django_cassandra_engine":
            return super(Command, self).handle_noargs(**options)
        else:
            return sync_cassandra.Command().execute(**options)

    def handle(self, **options):
        engine = get_engine_from_db_alias(options["database"])

        # Call regular syncdb if engine is different from ours
        if engine != "django_cassandra_engine":
            return super(Command, self).handle(**options)
        else:
            return sync_cassandra.Command().execute(**options)



from django.core.exceptions import NON_FIELD_ERRORS, FieldDoesNotExist, ValidationError

__ALL__ = (
    "_get_pk_val",
    "serializable_value",
    "full_clean",
    "_get_unique_checks",
    "_perform_unique_checks",
    "_perform_date_checks",
    "validate_unique",
    "clean",
    "clean_fields",
)


def _get_pk_val(self, meta=None):
    if not meta:
        meta = self._meta
    return getattr(self, meta.pk.attname)


def serializable_value(self, field_name):
    try:
        field = self._meta.get_field(field_name)
    except FieldDoesNotExist:
        return getattr(self, field_name)
    return getattr(self, field.attname)


def full_clean(self, exclude=None, validate_unique=True):
    # Taken from django.db.models.base
    errors = {}
    if exclude is None:
        exclude = []
    else:
        exclude = list(exclude)

    try:
        self.clean_fields(exclude=exclude)
    except ValidationError as e:
        errors = e.update_error_dict(errors)

    # Form.clean() is run even if other validation fails, so do the
    # same with Model.clean() for consistency.
    try:
        self.clean()
    except ValidationError as e:
        errors = e.update_error_dict(errors)

    # Run unique checks, but only for fields that passed validation.
    if validate_unique:
        for name in errors.keys():
            if name != NON_FIELD_ERRORS and name not in exclude:
                exclude.append(name)
        try:
            self.validate_unique(exclude=exclude)
        except ValidationError as e:
            errors = e.update_error_dict(errors)

    if errors:
        raise ValidationError(errors)


def _get_unique_checks(self, exclude=None):
    # Taken from django.db.models.base
    if exclude is None:
        exclude = []
    unique_checks = []

    unique_togethers = [(self.__class__, self._meta.unique_together)]
    for parent_class in self._meta.get_parent_list():
        if parent_class._meta.unique_together:
            unique_togethers.append((parent_class, parent_class._meta.unique_together))

    for model_class, unique_together in unique_togethers:
        for check in unique_together:
            for name in check:
                # If this is an excluded field, don't add this check.
                if name in exclude:
                    break
            else:
                unique_checks.append((model_class, tuple(check)))

    # These are checks for the unique_for_<date/year/month>.
    date_checks = []

    # Gather a list of checks for fields declared as unique and add them to
    # the list of checks.

    fields_with_class = [(self.__class__, self._meta.local_fields)]
    for parent_class in self._meta.get_parent_list():
        fields_with_class.append((parent_class, parent_class._meta.local_fields))

    for model_class, fields in fields_with_class:
        for f in fields:
            name = f.name
            if name in exclude:
                continue
            if f.unique:
                unique_checks.append((model_class, (name,)))
            if f.unique_for_date and f.unique_for_date not in exclude:
                date_checks.append((model_class, "date", name, f.unique_for_date))
            if f.unique_for_year and f.unique_for_year not in exclude:
                date_checks.append((model_class, "year", name, f.unique_for_year))
            if f.unique_for_month and f.unique_for_month not in exclude:
                date_checks.append((model_class, "month", name, f.unique_for_month))
    return unique_checks, date_checks


def _perform_unique_checks(self, unique_checks):
    # Taken from django.db.models.base
    errors = {}

    for model_class, unique_check in unique_checks:
        # Try to look up an existing object with the same values as this
        # object's values for all the unique field.

        lookup_kwargs = {}
        for field_name in unique_check:
            f = self._meta.get_field(field_name)
            lookup_value = getattr(self, f.attname)
            if lookup_value is None:
                # no value, skip the lookup
                continue
            if f.primary_key and not self._state.adding:
                # no need to check for unique primary key when editing
                continue
            lookup_kwargs[str(field_name)] = lookup_value

        # some fields were skipped, no reason to do the check
        if len(unique_check) != len(lookup_kwargs):
            continue

        qs = model_class._default_manager.filter(**lookup_kwargs)

        # Exclude the current object from the query if we are editing an
        # instance (as opposed to creating a new one)
        # Note that we need to use the pk as defined by model_class, not
        # self.pk. These can be different fields because model inheritance
        # allows single model to have effectively multiple primary keys.
        # Refs #17615.
        model_class_pk = self._get_pk_val(model_class._meta)
        if not self._state.adding and model_class_pk is not None:
            qs = qs.exclude(pk=model_class_pk)
        if qs.exists():
            if len(unique_check) == 1:
                key = unique_check[0]
            else:
                key = NON_FIELD_ERRORS
            errors.setdefault(key, []).append(
                self.unique_error_message(model_class, unique_check)
            )

    return errors


def _perform_date_checks(self, date_checks):
    # Taken from django.db.models.base
    errors = {}
    for model_class, lookup_type, field, unique_for in date_checks:
        lookup_kwargs = {}
        # there's a ticket to add a date lookup, we can remove this special
        # case if that makes it's way in
        date = getattr(self, unique_for)
        if date is None:
            continue
        if lookup_type == "date":
            lookup_kwargs["%s__day" % unique_for] = date.day
            lookup_kwargs["%s__month" % unique_for] = date.month
            lookup_kwargs["%s__year" % unique_for] = date.year
        else:
            lookup_kwargs["%s__%s" % (unique_for, lookup_type)] = getattr(
                date, lookup_type
            )
        lookup_kwargs[field] = getattr(self, field)

        qs = model_class._default_manager.filter(**lookup_kwargs)
        # Exclude the current object from the query if we are editing an
        # instance (as opposed to creating a new one)
        if not self._state.adding and self.pk is not None:
            qs = qs.exclude(pk=self.pk)

        if qs.exists():
            errors.setdefault(field, []).append(
                self.date_error_message(lookup_type, field, unique_for)
            )
    return errors


def validate_unique(self, exclude=None):
    # Taken from django.db.models.base
    unique_checks, date_checks = self._get_unique_checks(exclude=exclude)

    errors = self._perform_unique_checks(unique_checks)
    date_errors = self._perform_date_checks(date_checks)
    for k, v in date_errors.items():
        errors.setdefault(k, []).extend(v)

    if errors:
        raise ValidationError(errors)


def clean(self):
    # Taken from django.db.models.base
    pass


def clean_fields(self, exclude=None):
    # Taken from django.db.models.base
    if exclude is None:
        exclude = []

    errors = {}
    for f in self._meta.fields:
        if f.name in exclude:
            continue
        # Skip validation for empty fields with blank=True. The developer
        # is responsible for making sure they have a valid value.
        raw_value = getattr(self, f.attname)
        if f.blank and raw_value in f.empty_values:
            continue
        try:
            setattr(self, f.attname, f.clean(raw_value, self))
        except ValidationError as e:
            errors[f.name] = e.error_list

    if errors:
        raise ValidationError(errors)



from django.core.management import call_command
from django.core.management.commands.migrate import Command as MigrateCommand

from django_cassandra_engine.utils import get_engine_from_db_alias


class Command(MigrateCommand):
    def handle(self, *args, **options):
        engine = get_engine_from_db_alias(options["database"])

        # Call regular migrate if engine is different from ours
        if engine != "django_cassandra_engine":
            return super(Command, self).handle(*args, **options)
        else:
            self.stdout.write(
                "Migrations are not supported in this engine. "
                "Calling syncdb instead.."
            )
            call_command("sync_cassandra", **options)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_cassandra_engine',
 'django_cassandra_engine.base',
 'django_cassandra_engine.management',
 'django_cassandra_engine.management.commands',
 'django_cassandra_engine.models',
 'django_cassandra_engine.rest',
 'django_cassandra_engine.sessions',
 'django_cassandra_engine.sessions.backends']

package_data = \
{'': ['*']}

install_requires = \
['django>=2.0,<5.0', 'scylla-driver>=3.25.0,<4.0.0']

setup_kwargs = {
    'name': 'django-cassandra-engine',
    'version': '1.7.0',
    'description': 'Django Cassandra Engine',
    'long_description': '\n# Django Cassandra Engine - the Cassandra backend for Django #\n\nAll tools you need to start your journey with Apache Cassandra and Django Framework!\n\n[![Latest version](https://img.shields.io/pypi/v/django-cassandra-engine.svg "Latest version")](https://pypi.python.org/pypi/django-cassandra-engine/)\n![workflow](https://github.com/r4fek/django-cassandra-engine/actions/workflows/tox.yml/badge.svg)\n\nDiscord: https://discord.gg/pxunMGmDNc\n## Features ##\n\n* integration with latest `python-driver` and optionally `dse-driver` from DataStax\n* working `flush`, `migrate`, `sync_cassandra`, `inspectdb` and\n  `dbshell` commands\n* support for creating/destroying test database\n* accepts all `Cqlengine` and `cassandra.cluster.Cluster` connection options\n* automatic connection/disconnection handling\n* works well along with relational databases (as secondary DB)\n* storing sessions in Cassandra\n* working django forms\n* usable admin panel with Cassandra models\n* support DataStax Astra cloud hosted Cassandra\n\n## Sponsors ##\nHelp support ongoing development and maintenance by [sponsoring Django Cassandra Engine](https://github.com/sponsors/r4fek).\n\n### Our Sponsors: ###\n<table><tr>\n<td align="center" width="300" ><a href="https://astra.dev/3xPljcu"><img src="https://www.datastax.com/sites/default/files/2021-07/astra-negative-square.png" width="90" height="90" alt="Astra DB" /><br />Astra DB</a><br/>Use Django with DataStax Astra DB - built on Apache Cassandra.</td>\n<td align="center" width="300" ><a href="https://github.com/NoiSek"><img src="https://avatars.githubusercontent.com/u/631328?v=4" width="90" height="90" alt="NoiSek" /><br/>NoiSek</a></td>\n</tr></table>\n\n\n## Installation ##\n\nRecommended installation:\n\n    pip install django-cassandra-engine\n\n## Basic Usage ##\n\n1. Add `django_cassandra_engine` to `INSTALLED_APPS` in your `settings.py` file:\n\n        INSTALLED_APPS = (\'django_cassandra_engine\',) + INSTALLED_APPS\n\n2. Change `DATABASES` setting:\n\n        DATABASES = {\n            \'default\': {\n                \'ENGINE\': \'django_cassandra_engine\',\n                \'NAME\': \'db\',\n                \'TEST_NAME\': \'test_db\',\n                \'HOST\': \'db1.example.com,db2.example.com\',\n                \'OPTIONS\': {\n                    \'replication\': {\n                        \'strategy_class\': \'SimpleStrategy\',\n                        \'replication_factor\': 1\n                    }\n                }\n            }\n        }\n\n3. Define some model:\n\n        # myapp/models.py\n\n        import uuid\n        from cassandra.cqlengine import columns\n        from django_cassandra_engine.models import DjangoCassandraModel\n\n        class ExampleModel(DjangoCassandraModel):\n            example_id    = columns.UUID(primary_key=True, default=uuid.uuid4)\n            example_type  = columns.Integer(index=True)\n            created_at    = columns.DateTime()\n            description   = columns.Text(required=False)\n\n4. Run `./manage.py sync_cassandra`\n5. Done!\n\n## Connect to Cassandra with a Cloud Config bundle ##\nTo connect to a hosted Cassandra cluster that provides a secure connection bundle (ex. DataStax Astra) change the `DATABASES` setting of your settings.py:\n\n        DATABASES = {\n            \'default\': {\n                \'ENGINE\': \'django_cassandra_engine\',\n                \'NAME\': \'keyspace_name\',\n                \'TEST_NAME\': \'table_name\',\n                \'USER\': \'token\',\n                \'PASSWORD\': token_value,\n                \'OPTIONS\': {\n                    \'connection\': {\n                        \'cloud\': {\n                            \'secure_connect_bundle\': \'/path/to/secure/bundle.zip\'\n                        },\n                    }\n                }\n            }\n        }\n\n## Documentation ##\n\nThe documentation can be found online [here](http://r4fek.github.io/django-cassandra-engine/).\n\n## License ##\nCopyright (c) 2014-2022, [Rafał Furmański](https://linkedin.com/in/furmanski).\n\nAll rights reserved. Licensed under BSD 2-Clause License.\n',
    'author': 'Rafał Furmański',
    'author_email': 'r.furmanski@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/r4fek/django-cassandra-engine',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
